import base64
import json
from time import time

import requests
from requests.auth import AuthBase

from pyakeneo.utils import urljoin


class Auth(AuthBase):
    TOKEN_PATH = "api/oauth/v1/token"
    TOKEN_EXPIRY_SECURITY = 60 * 2

    def __init__(self, base_url, client_id, secret, username, password):
        """
        :param base_url: eg http://localhost:8088/
        """
        self._base_url = base_url
        self._client_id = client_id
        self._secret = secret
        self._username = username
        self._password = password
        self._token = None
        self._refresh_token = None
        self._expiry_date = None

    @property
    def authorization(self):
        return "Bearer {0}".format(self._token)

    def _request_a_token(self, grant_type="password"):
        """Requests a token. Throws in case of error"""
        authorization = "Basic {0}".format(
            base64.b64encode(
                "{0}:{1}".format(self._client_id, self._secret).encode("ascii")
            ).decode("utf-8")
        )
        headers = {"Content-Type": "application/json", "Authorization": authorization}
        if grant_type == "refresh_token" and self._refresh_token:
            data = json.dumps(
                {"grant_type": "refresh_token", "refresh_token": self._refresh_token}
            )
        elif grant_type == "password" or (
            grant_type == "refresh_token" and not self._refresh_token
        ):
            data = json.dumps(
                {
                    "grant_type": "password",
                    "username": self._username,
                    "password": self._password,
                }
            )
        else:
            raise ValueError(
                "grant_type parameter is expected to be either "
                + '"password" or "refresh_token". {0} provided'.format(grant_type)
            )

        url = urljoin(self._base_url, self.TOKEN_PATH)
        r = requests.post(url, data=data, headers=headers)
        r.raise_for_status()

        try:
            json_data = json.loads(r.text)
        except json.decoder.JSONDecodeError as e:
            raise SyntaxError(
                "The server did not return expected json: {0}".format(r.text)
            )

        try:
            self._token = json_data["access_token"]
            self._refresh_token = json_data["refresh_token"]
        except KeyError:
            raise SyntaxError(
                "The server did not return expected tokens: {0}".format(json_data)
            )

        try:
            self._expiry_date = time() + float(json_data["expires_in"])
        except KeyError:
            self._expiry_date = None
        except ValueError:
            raise SyntaxError(
                "The server did not return a valid expires_in: {0}".format(json_data)
            )

    def _should_refresh_token(self):
        """Returns True if the token is expired / about to expire"""
        if not self._expiry_date:
            return True
        else:
            return time() > self._expiry_date - self.TOKEN_EXPIRY_SECURITY

    def _refresh_the_token(self):
        """Requests a new token based on refresh token."""
        return self._request_a_token(grant_type="refresh_token")

    def __call__(self, r):
        if not self._token or not self._refresh_token:
            self._request_a_token()

        if self._should_refresh_token():
            self._refresh_the_token()

        r.headers["Authorization"] = self.authorization
        return r



from __future__ import annotations

import json
from typing import Dict, Iterable

import requests


class Result(object):
    """
    Holds the result of a search. It can be iterated through as a list,
    as an iterator, or as a generator. Note that only one iteration should be
    used on a given Collection object.
    Search results are paginated: https://api.akeneo.com/documentation/pagination.html

    The next page will be loaded once the user iterated over the whole current page.
    The content of the new page will replace the content of the previous page.
    """

    def __init__(
        self,
        session: requests.Session,
        *,
        items: list | dict,
        count: int,
        link_first: str,
        link_next: str,
        link_self: str,
    ):
        self._session = session
        self._items = items
        self._count = count
        self._link_next = link_next
        self._link_self = link_self
        self._link_first = link_first

        self._page_iterator = iter(self._items)
        self._reached_the_end = False

    def __iter__(self):
        while not self._reached_the_end:
            for item in self._page_iterator:
                yield item
            self.fetch_next_page()

    def __next__(self):
        try:
            return next(self._page_iterator)
        except StopIteration:
            self.fetch_next_page()
            if not self._reached_the_end:
                return next(self._page_iterator)
            else:
                return

    def get_page_items(self):
        return self._items

    def fetch_next_page(self):
        """Return True if a next page exists. Returns False otherwise."""
        if self._link_next:
            response = self._session.get(self._link_next)
            if response.ok:
                next_page = Result.parse_page(json.loads(response.text))
                self._items = next_page["items"]
                self._count = next_page["count"]
                self._link_next = next_page["link_next"]
                self._link_self = next_page["link_self"]
                self._link_first = next_page["link_first"]

                self._page_iterator = iter(self._items)
                self._reached_the_end = False
            else:
                self._reached_the_end = True
        else:
            self._reached_the_end = True
        return not self._reached_the_end

    def get_count(self):
        return self._count

    def get_next_link(self):
        return self._link_next

    def get_self_link(self):
        return self._link_self

    def get_first_link(self):
        return self._link_first

    @classmethod
    def parse_page(cls, json_data: dict) -> Dict:
        """Returns (next link, retrieved items, count of items)"""
        final_next_link = None
        next_link = json_data["_links"].get("next")
        if next_link:
            final_next_link = next_link["href"]
        return {
            "items": json_data["_embedded"]["items"],
            "count": json_data.get("items_count"),
            "link_first": json_data["_links"]["first"]["href"],
            "link_next": final_next_link,
            "link_self": json_data["_links"]["self"]["href"],
        }

    @classmethod
    def parse_result(cls, session: requests.Session, json_data: dict | list):
        if cls.is_paginated(json_data):
            return cls(session, **cls.parse_page(json_data))
        else:
            return cls(session, **cls.parse_non_paginated(json_data))

    @staticmethod
    def from_json_text(session: requests.Session, json_text: str) -> "Result":
        json_data = json.loads(json_text)
        return Result.parse_result(session, json_data)

    @classmethod
    def is_paginated(cls, json_data: dict | list):
        return isinstance(json_data, dict) and "_links" in json_data

    @classmethod
    def parse_non_paginated(cls, json_data: list) -> Dict:
        link_self = ""
        if json_data and "_links" in json_data[0]:
            link_self = json_data["_links"]["self"]["href"]
        return {
            "items": json_data,
            "count": len(json_data),
            "link_first": "",
            "link_next": "",
            "link_self": link_self,
        }



import json
import math

import requests

from pyakeneo import interfaces
from pyakeneo.result import Result
from pyakeneo.utils import urljoin


class CreatableResource(interfaces.CreatableResourceInterface):
    def create_item(self, item):
        url = self._endpoint
        r = self._session.post(url, data=json.dumps(item, separators=(",", ":")))

        if r.status_code != 201:
            raise requests.HTTPError(
                "Status code: {0}. Content: {1}".format(r.status_code, r.text)
            )


class ListableResource(interfaces.ListableResourceInterface):
    def fetch_list(self, args=None):
        """Send a request with search, etc.
        Returns an iterable list (Collection)"""
        if args:
            for (key, value) in args.items():
                if not isinstance(value, str):
                    args[key] = json.dumps(value)

        url = self._endpoint
        r = self._session.get(url, params=args)
        r.raise_for_status()

        c = Result.from_json_text(self._session, json_text=r.text)
        return c


class SearchAfterListableResource(ListableResource):
    def fetch_list(self, args=None):
        """Send a request with search, etc.
        Returns an iterable list (Collection)"""
        params = args
        if not params:
            params = {"pagination_type": "search_after"}
        elif "pagination_type" not in params:
            params["pagination_type"] = "search_after"

        return super(SearchAfterListableResource, self).fetch_list(params)


class GettableResource(interfaces.GettableResourceInterface):
    def fetch_item(self, code_or_item):
        """Returns a unique item object. code_or_item should be the code
        of the desired item, or an item with the proper code."""
        code = code_or_item
        if not isinstance(code_or_item, str):
            # if code_or_item is item, then fetch the code
            code = self.get_code(code_or_item)

        url = urljoin(self._endpoint, code)
        r = self._session.get(url)
        r.raise_for_status()

        return json.loads(r.text)  # returns item as a dict


class DeletableResource(interfaces.DeletableResourceInterface):
    def delete_item(self, code_or_item):
        """code_or_item should be the code
        of the desired item, or an item with the proper code."""
        code = code_or_item
        if not isinstance(code_or_item, str):
            # if code_or_item is item, then fetch the code
            code = self.get_code(code_or_item)
        url = urljoin(self._endpoint, code)
        r = self._session.delete(url)

        if r.status_code != 204:
            raise requests.HTTPError(
                "The item {0} doesn't exit. Content: {1}".format(code, r.text)
            )


class UpdatableResource(interfaces.UpdatableResourceInterface):
    def update_create_item(self, item_values, code=None):
        if not code:
            code = self.get_code(item_values)

        url = urljoin(self._endpoint, code)
        r = self._session.patch(
            url, data=json.dumps(item_values, separators=(",", ":"))
        )

        if r.status_code not in [201, 204]:
            raise requests.HTTPError(
                "Status code: {0}. Content: {1}".format(r.status_code, r.text)
            )
        else:
            return r.headers.get("Location")


class UpdatableListResource(interfaces.UpdatableResourceInterface):
    def update_create_list(self, items, code=None):
        url = self._endpoint
        data = ""
        for item in items:
            data += json.dumps(item, separators=(",", ":")) + "\n"
        r = self._session.patch(
            url,
            data=data,
            headers={"Content-type": "application/vnd.akeneo.collection+json"},
        )

        if r.status_code == 413:
            # TODO handle 413
            # Request Entity Too Large
            # There are too many resources to process (max 100)
            # =>>>>> or the line of JSON is too long (max 1 000 000 characters)
            # split the list in several chunks
            num = 100
            n = math.ceil(len(items) / num)

            itemss = [items[i: i + num] for i in range(0, (n - 1) * num, num)]
            itemss.append(items[(n - 1) * num:])

            return [
                item
                for those_items in itemss
                for item in self.update_create_list(those_items)
            ]

        if r.status_code != 200:
            r.raise_for_status()

        else:
            statuses = []
            for line in r.text.split("\n"):
                statuses.append(json.loads(line))
            return statuses


class IdentifierBasedResource(interfaces.CodeBasedResourceInterface):
    def get_code(self, item):
        return item["identifier"]


class CodeBasedResource(interfaces.CodeBasedResourceInterface):
    def get_code(self, item):
        return item["code"]


class EnterpriseEditionResource:
    pass


class ResourcePool:
    def __init__(self, endpoint, session):
        """Initialize the ResourcePool to the given endpoint. Eg: products"""
        self._endpoint = endpoint
        self._session = session

    def get_url(self):
        return self._endpoint


class ProductsPool(
    ResourcePool,
    IdentifierBasedResource,
    CreatableResource,
    DeletableResource,
    GettableResource,
    SearchAfterListableResource,
    UpdatableResource,
    UpdatableListResource,
):
    """https://api.akeneo.com/api-reference.html#Products"""

    # TODO: EE support of drafts
    pass


class ProductModelsPool(
    ResourcePool,
    IdentifierBasedResource,
    CreatableResource,
    GettableResource,
    SearchAfterListableResource,
    UpdatableResource,
):
    """https://api.akeneo.com/api-reference.html#Productmodel"""

    pass


class PublishedProductsPool(
    ResourcePool,
    IdentifierBasedResource,
    GettableResource,
    SearchAfterListableResource,
    EnterpriseEditionResource,
):
    """https://api.akeneo.com/api-reference.html#Publishedproduct"""

    pass


class CategoriesPool(
    ResourcePool,
    CodeBasedResource,
    CreatableResource,
    GettableResource,
    ListableResource,
    UpdatableResource,
    UpdatableListResource,
):
    """https://api.akeneo.com/api-reference.html#Category"""

    pass


class FamilyVariantsPool(
    ResourcePool,
    CodeBasedResource,
    CreatableResource,
    GettableResource,
    ListableResource,
):
    """https://api.akeneo.com/api-reference.html#Familyvariant"""

    pass


class FamiliesPool(
    ResourcePool,
    CodeBasedResource,
    CreatableResource,
    DeletableResource,
    GettableResource,
    ListableResource,
    UpdatableResource,
    UpdatableListResource,
):
    """https://api.akeneo.com/api-reference.html#Family"""

    def variants(self, code):
        return FamilyVariantsPool(
            urljoin(self._endpoint, code, "variants/"), self._session
        )


class AttributeOptionsPool(
    ResourcePool,
    CodeBasedResource,
    CreatableResource,
    GettableResource,
    ListableResource,
    UpdatableResource,
):
    """https://api.akeneo.com/api-reference.html#Attributeoptions"""

    pass


class AttributesPool(
    ResourcePool,
    CodeBasedResource,
    CreatableResource,
    GettableResource,
    ListableResource,
    UpdatableResource,
    UpdatableListResource,
):
    """https://api.akeneo.com/api-reference.html#Attributes"""

    def options(self, code):
        return AttributeOptionsPool(
            urljoin(self._endpoint, code, "options/"), self._session
        )


class AttributeGroupsPool(
    ResourcePool,
    CodeBasedResource,
    ListableResource,
    CreatableResource,
    UpdatableListResource,
    GettableResource,
    UpdatableResource,
):
    """https://api.akeneo.com/api-reference.html#Attributegroups"""

    pass


class MediaFilesPool(
    ResourcePool,
    CodeBasedResource,
    ListableResource,
    CreatableResource,
    GettableResource,
):
    """https://api.akeneo.com/api-reference.html#Mediafiles"""

    def download(self, code):
        # TODO: implement this method
        raise NotImplementedError()


class LocalesPool(
    ResourcePool,
    CodeBasedResource,
    ListableResource,
    GettableResource,
):
    """https://api.akeneo.com/api-reference.html#Locales"""

    pass


class ChannelsPool(
    ResourcePool,
    CodeBasedResource,
    ListableResource,
    UpdatableListResource,
    GettableResource,
    UpdatableResource,
):
    """https://api.akeneo.com/api-reference.html#Channels"""

    pass


class CurrenciesPool(
    ResourcePool,
    CodeBasedResource,
    ListableResource,
    CreatableResource,
):
    """https://api.akeneo.com/api-reference.html#Currencies"""

    pass


class MeasureFamiliesPool(
    ResourcePool,
    CodeBasedResource,
    ListableResource,
    GettableResource,
):
    """https://api.akeneo.com/api-reference.html#Measurefamilies"""

    pass


class AssociationTypesPool(
    ResourcePool,
    CodeBasedResource,
    ListableResource,
    CreatableResource,
    UpdatableListResource,
    GettableResource,
    UpdatableResource,
):
    """https://api.akeneo.com/api-reference.html#Associationtypes"""

    pass


class AssetsPool(
    ResourcePool,
    CodeBasedResource,
    GettableResource,
    ListableResource,
    UpdatableResource,
    UpdatableListResource,
    DeletableResource,
):
    pass


class AssetFamilyPool(
    ResourcePool,
    CodeBasedResource,
    GettableResource,
    ListableResource,
    UpdatableResource,
):
    def assets(self, code):
        return AssetsPool(urljoin(self._endpoint, code, "assets/"), self._session)


class ReferenceEntityRecordPool(
    ResourcePool,
    CodeBasedResource,
    GettableResource,
    ListableResource,
    UpdatableResource,
    UpdatableListResource,
):
    pass


class ReferenceEntityAttributeOptionsPool(
    ResourcePool,
    CodeBasedResource,
    GettableResource,
    ListableResource,
    UpdatableResource,
):
    pass


class ReferenceEntityAttributePool(
    ResourcePool,
    CodeBasedResource,
    GettableResource,
    ListableResource,
    UpdatableResource,
):
    def options(self, code):
        return ReferenceEntityAttributeOptionsPool(
            urljoin(self._endpoint, code, "options/"), self._session
        )


class ReferenceEntityPool(
    ResourcePool,
    CodeBasedResource,
    GettableResource,
    ListableResource,
    UpdatableResource,
):
    def records(self, entity_code):
        return ReferenceEntityRecordPool(
            urljoin(self._endpoint, entity_code, "records/"), self._session
        )

    def attributes(self, entity_code):
        return ReferenceEntityAttributePool(
            urljoin(self._endpoint, entity_code, "attributes/"), self._session
        )



import json
from collections import namedtuple


def urljoin(*args):
    """
    Joins given arguments into a url. Trailing but not leading slashes are
    stripped for each argument.
    https://stackoverflow.com/a/11326230
    """
    return "/".join(map(lambda x: str(x).strip("/").rstrip("/"), args))


def _json_object_hook(data):
    """https://stackoverflow.com/a/15882054"""
    try:
        data["links"] = data.pop("_links")
    except KeyError as e:
        pass
    try:
        data["embedded"] = data.pop("_embedded")
    except KeyError as e:
        pass
    return namedtuple("X", data.keys(), rename=False)(*data.values())


def json2object(data):
    """https://stackoverflow.com/a/15882054"""
    return json.loads(data, object_hook=_json_object_hook)
